/**
 * This is the file that handles insertion into a MongoDB collection. There are three
 * primary steps to this process.
 * First, the collection in MongoDB is queried to determine
 * how many days' worth of data need to be obtained. If this number is 0, nothing will happen.
 * Else, the function will grab however many days' worth of data is necessary from the streams,
 * but only that much data so as to preserve memory.
 *
 * Second, the program uses async.waterfall to asynchronously grab the data from
 * each of the different metrics so that they can be dealt with synchronously. Each
 * metric grabs data and puts it in the form such that each day is in an array - each
 * object across all metrics has the same format:
 * {
 * date:
 * name:
 * metrics:
 * stats:
 * }
 * where date is the relevant date, name is the name of the metric, metrics is a data
 * dump from the stream, and stats are any relevant statistics from the raw data dump.
 *
 * Third, once all the data is passed down through waterfall as arrays of objects, the arrays
 * are consolidated into a centralized object where each day contains data for all metrics
 * on that day. This object is then transformed into an array to be used as a parameter in
 * the insertMany function. The data is then inserted into the database.
*/

/* eslint no-console:0 */
/* eslint guard-for-in:0 */
/* eslint-env es6 */

require('dotenv').config();
var mongo = require('mongodb').MongoClient;
var assert = require('assert');
var compassUsageReport = require('../../');
var async = require('async');
var helpers = require('../../lib/helpers.js');

/**
 * Throws an error if .env file is not configured properly.
*/
helpers.verifyEnv();

/**
 * First function in the waterfall. Grabs OS launches data by day, and puts
 * the data into the correct format of an array of objects. Passes all relevant
 * data and connection info through the callback.
 * @param {db} db                     db connection established by node driver
 * @param {string} coll               collection to write into
 * @param {int} numDays               number of days of data to grab
 * @param {function} callback         callback function passed through the waterfall
 * @returns {null}                    no return - callback
*/
function getOSLaunchesData(db, coll, numDays, callback) {
  if (numDays === 0) {
    return callback(null, null, db, coll, numDays);
  }
  var stream = compassUsageReport.getLaunchesOsData(process.env.MIXPANEL_API_SECRET);
  var osLaunchesDataArray = [];
  console.log('Getting OS launches data.');
  stream.on('data', function(data) {
    if (osLaunchesDataArray.length === numDays) {
      return;
    }
    osLaunchesDataArray.push({
      date: data.date,
      name: 'osLaunches',
      metrics: data.values,
      stats: []
    });
  });
  stream.on('end', function() {
    callback(null, osLaunchesDataArray, db, coll, numDays);
  });
}

/**
 * Second function in the waterfall. Grabs version launches by day, and puts the
 * data into the correct format of an array of objects. Replaces '.' in versions
 * with '_' so that they can be keys in MongoDB. Passes all relevant
 * data and connection info through the callback.
 * @param {array} osLaunches          array of objects for osLaunches data passed through waterfall
 * @param {db} db                     db connection established by node driver
 * @param {string} coll               collection to write into
 * @param {int} numDays               number of days of data to grab
 * @param {function} callback         callback function passed through the waterfall
 * @returns {null}                    no return - callback
*/
function getVersionLaunchesData(osLaunches, db, coll, numDays, callback) {
  if (numDays === 0) {
    return callback(null, osLaunches, null, db, coll, numDays);
  }
  var stream = compassUsageReport.getLaunchesVersionData(process.env.MIXPANEL_API_SECRET);
  var versionLaunchesDataArray = [];
  console.log('Getting version launches data.');
  stream.on('data', function(data) {
    if (versionLaunchesDataArray.length === numDays) {
      return;
    }
    var dataObj = {
      date: data.date,
      name: 'versionLaunches',
      stats: []
    };
    var values = {};
    for (var version in data.values) {
      if (data.values.hasOwnProperty(version)) {
        values[version.replace(/\./g, '_')] = data.values[version];
      }
    }
    dataObj.metrics = values;
    versionLaunchesDataArray.push(dataObj);
  });
  stream.on('end', function() {
    callback(null, osLaunches, versionLaunchesDataArray, db, coll, numDays);
  });
}

/**
 * Third function in the waterfall. Grabs session data per day per user and
 * compiles the data into the correct format. For the stats property of the object,
 * this function creates an array of session lengths and also an array of number of
 * sessions by iterating over the users then computes the relevant statistics. Passes
 * along all relevant data and connection info through the callback.
 * @param {array} osLaunches          array of objects for osLaunches data passed through waterfall
 * @param {array} versionLaunches     array of objects for versionLaunches data passed through waterfall
 * @param {db} db                     db connection established by node driver
 * @param {string} coll               collection to write into
 * @param {int} numDays               number of days of data to grab
 * @param {function} callback         callback function passed through the waterfall
 * @returns {null}                    no return - callback
*/
function getSessionData(osLaunches, versionLaunches, db, coll, numDays, callback) {
  if (numDays === 0) {
    return callback(null, osLaunches, versionLaunches, null, db, coll, numDays);
  }
  var stream = compassUsageReport.getSessionData(process.env.MIXPANEL_API_SECRET);
  var sessionDataArray = [];
  console.log('Getting session data');
  stream.on('data', function(data) {
    if (sessionDataArray.length === numDays) {
      return;
    }
    var dataObj = {
      date: data.date,
      name: 'session',
      metrics: data.values
    };
    var sessions = [];
    var lengths = [];
    var numUsers = 0;
    for (var user in data.values) {
      if (data.values.hasOwnProperty(user)) {
        numUsers++;
        sessions.push(data.values[user].sessions);
        lengths.push(data.values[user].lengths);
      }
    }
    var stats = {
      numUsers: numUsers,
      numSessions: helpers.getStats(sessions),
      sessionLengths: helpers.getStats(lengths)
    };
    dataObj.stats = stats;
    sessionDataArray.push(dataObj);
  });
  stream.on('end', function() {
    callback(null, osLaunches, versionLaunches, sessionDataArray, db, coll, numDays);
  });
}

/**
 * Fourth function in the waterfall. Grabs sampling data by day and compiles it into
 * the appropriate format. Since the sampling data comes in as an array for each day,
 * the stats portion of this object is just the result of passing the array to getStats().
 * Passes all relevant data and connection info through the callback.
 * @param {array} osLaunches          array of objects for osLaunches data passed through waterfall
 * @param {array} versionLaunches     array of objects for versionLaunches data passed through waterfall
 * @param {array} session             array of objects for session data passed through waterfall
 * @param {db} db                     db connection established by node driver
 * @param {string} coll               collection to write into
 * @param {int} numDays               number of days of data to grab
 * @param {function} callback         callback function passed through the waterfall
 * @returns {null}                    no return - callback
*/
function getSamplingData(osLaunches, versionLaunches, session, db, coll, numDays, callback) {
  if (numDays === 0) {
    return callback(null, osLaunches, versionLaunches, session, null, db, coll, numDays);
  }
  var samplingDataArray = [];
  console.log('Getting sampling data.');
  var stream = compassUsageReport.getSamplingData(process.env.MIXPANEL_API_SECRET);
  stream.on('data', function(data) {
    if (samplingDataArray.length === numDays) {
      return;
    }
    samplingDataArray.push({
      date: data.date,
      name: 'sampling',
      metrics: data.values,
      stats: {
        numSamples: data.values.length,
        sampling: helpers.getStats(data.values)
      }
    });
  });
  stream.on('end', function() {
    callback(null, osLaunches, versionLaunches, session, samplingDataArray, db, coll, numDays);
  });
}

/**
 * Fifth function in the waterfall. Grabs instance connection flow data and manipulates
 * the data into the correct form. Passes the data and connection info along through the
 * callback.
 * @param {array} osLaunches          array of objects for osLaunches data passed through waterfall
 * @param {array} versionLaunches     array of objects for versionLaunches data passed through waterfall
 * @param {array} session             array of objects for session data passed through waterfall
 * @param {array} sampling            array of objects for sampling data passed through waterfall
 * @param {db} db                     db connection established by node driver
 * @param {string} coll               collection to write into
 * @param {int} numDays               number of days of data to grab
 * @param {function} callback         callback function passed through the waterfall
 * @returns {null}                    no return - callback
*/
function getInstanceData(osLaunches, versionLaunches, session, sampling, db, coll, numDays, callback) {
  if (numDays === 0) {
    return callback(null, osLaunches, versionLaunches, session, sampling, null, db, coll, numDays);
  }
  var instanceConnectionDataArray = [];
  console.log('Getting instance connection data.');
  var stream = compassUsageReport.getInstanceData(process.env.MIXPANEL_API_SECRET);
  stream.on('data', function(data) {
    if (instanceConnectionDataArray.length === numDays) {
      return;
    }
    var dataObj = {
      date: data.date,
      name: 'instance',
      metrics: data.values
    };
    var stats = {
      starting_amount: data.values.analysis.starting_amount,
      completion: data.values.analysis.completion,
      conversion: data.values.analysis.completion / data.values.analysis.starting_amount
    };
    dataObj.stats = stats;
    instanceConnectionDataArray.push(dataObj);
  });
  stream.on('end', function() {
    callback(null, osLaunches, versionLaunches, session, sampling, instanceConnectionDataArray, db, coll, numDays);
  });
}

/**
 * Sixth function in the waterfall. Grabs successful update flow data and manipulates
 * the data into the correct form. Passes the data and connection info along through the
 * callback.
 * @param {array} osLaunches          array of objects for osLaunches data passed through waterfall
 * @param {array} versionLaunches     array of objects for versionLaunches data passed through waterfall
 * @param {array} session             array of objects for session data passed through waterfall
 * @param {array} sampling            array of objects for sampling data passed through waterfall
 * @param {array} instance            array of objects for instance connection flow data passed through waterfall
 * @param {db} db                     db connection established by node driver
 * @param {string} coll               collection to write into
 * @param {int} numDays               number of days of data to grab
 * @param {function} callback         callback function passed through the waterfall
 * @returns {null}                    no return - callback
*/
function getUpdateData(osLaunches, versionLaunches, session, sampling, instance, db, coll, numDays, callback) {
  if (numDays === 0) {
    return callback(null, osLaunches, versionLaunches, session, sampling, instance, null, db, coll, numDays);
  }
  var updateDataArray = [];
  console.log('Getting update success data.');
  var stream = compassUsageReport.updateSuccessData(process.env.MIXPANEL_API_SECRET);
  stream.on('data', function(data) {
    if (updateDataArray.length === numDays) {
      return;
    }
    var dataObj = {
      date: data.date,
      name: 'update success',
      metrics: data.values
    };
    var stats = {
      starting_amount: data.values.analysis.starting_amount,
      completion: data.values.analysis.completion,
      conversion: data.values.analysis.completion / data.values.analysis.starting_amount
    };
    dataObj.stats = stats;
    updateDataArray.push(dataObj);
  });
  stream.on('end', function() {
    callback(null, osLaunches, versionLaunches, session, sampling, instance, updateDataArray, db, coll, numDays);
  });
}

/**
 * Function that essentially takes all of the arrays of objects and creates a single object
 * that has data for each day for each metric. Makes use of the arguments object, since each array
 * is passed as an argument via the waterfall callback. Passes the object along through the callback.
 * Object is of the form:
 * {
 *    '2016-08-02' : {
 *                      metrics: {
 *                                  osLaunches: osLaunches metrics,
 *                                  versionLaunches: versionLaunches metrics....
 *                              },
 *                      stats: {
 *                                  osLaunches: osLaunches stats,
 *                                  versionLaunches: versionLaunches stats
 *                                }
 *                    },
 *    '2016-08-01' : {
 *                      metrics: {
 *                                  osLaunches: osLaunches metrics,
 *                                  versionLaunches: versionLaunches metrics....
 *                              },
 *                      stats: {
 *                                  osLaunches: osLaunches stats,
 *                                  versionLaunches: versionLaunches stats
 *                                }
 *                    }
 *  }
 * @param {array} osLaunches          array of objects for osLaunches data passed through waterfall
 * @param {array} versionLaunches     array of objects for versionLaunches data passed through waterfall
 * @param {array} session             array of objects for session data passed through waterfall
 * @param {array} sampling            array of objects for sampling data passed through waterfall
 * @param {array} instance            array of objects for instance connection flow data passed through waterfall
 * @param {array} update              array of objects for successful update flow data passed through waterfall
 * @param {db} db                     db connection established by node driver
 * @param {string} coll               collection to write into
 * @param {int} numDays               number of days of data to grab
 * @param {function} callback         callback function passed through the waterfall
 * @returns {null}                    no return - callback
*/
function writeObj(osLaunches, versionLaunches, session, sampling, instance, update, db, coll, numDays, callback) {
  if (numDays === 0) {
    return callback(null, null, db, coll, numDays);
  }
  var mongoObj = {};
  var len;
  console.log('Merging data.');
  for (var i = 0; i < arguments.length - 3; ++i) {
    len = arguments[i].length;
    for (var j = 0; j < len; ++j) {
      if (mongoObj[arguments[i][j].date] === undefined) {
        mongoObj[arguments[i][j].date] = {
          stats: {},
          metrics: {}
        };
      }
      mongoObj[arguments[i][j].date].stats[arguments[i][j].name] = arguments[i][j].stats;
      mongoObj[arguments[i][j].date].metrics[arguments[i][j].name] = arguments[i][j].metrics;
    }
  }
  callback(null, mongoObj, db, coll, numDays);
}

/**
 * Function that writes all of the data to MongoDB. First, the function takes
 * the object created in writeObj and creates an array of objects, where each object
 * is an entry to the database. Thus, this array is passed directly to insertMany.
 * The function uses the database connection passed all of the way through the
 * waterfall in order to make the insert.
 * @param {obj} mongoObj             object with data to be written to database
 * @param {db} db               db connection
 * @param {string} coll         collection to write to
 * @param {int} numDays         number of days of data to write
 * @param {function} callback   callback function passed through the waterfall
 * @returns {null}              no return - callback
*/
function writeToDB(mongoObj, db, coll, numDays, callback) {
  if (numDays === 0) {
    db.close();
    return callback(null, 'no data to insert');
  }
  var insertManyArray = [];
  for (var date in mongoObj) {
    var dayObj = {};
    if (mongoObj.hasOwnProperty(date)) {
      dayObj._id = date;
      dayObj.metrics = mongoObj[date].metrics;
      dayObj.stats = mongoObj[date].stats;
    }
    insertManyArray.push(dayObj);
  }
  db.collection(coll).insertMany(insertManyArray, function(err, r) {
    assert.equal(null, err);
    assert.equal(numDays, r.insertedCount);
    console.log('Data successfully inserted.');
    db.close();
    console.log('Database has closed!');
    callback(null, 'done');
  });
}

/**
 * Function that executes the waterfall. The first function in the waterfall,
 * launchesByOS() has to be wrapped in another function that has the callback
 * because in the waterfall function, the first function cannot have more than
 * one argument, and the database connection info has to be passed through. Thus,
 * the function it's wrapped in creates the database connection, queries MongoDB
 * to determine how many days' worth of data are necessary, then calls the waterfall.
 * @param {array} argv          command line arguments - used to build connection url
*/
function exportToMongo(argv) {
  async.waterfall([
    function(callback) {
      var uri = helpers.getURL(argv.url, argv.db);
      console.log(`Connecting to database with uri ${argv.url}.`);
      mongo.connect(uri, function(err, db) {
        assert.equal(null, err);
        console.log('Connected succesfully to server');
        console.log('Computing number of days to insert into database');
        db.collection(argv.coll).find({}, { 'sort': [['_id', 'desc']], 'limit': 1}).toArray(function(error, docs) {
          assert.equal(error, null);
          var numDays = 30;
          if (docs.length) {
            var mostRecentDate = docs[0]._id;
            numDays = helpers.daysBetween(mostRecentDate, helpers.formatDate(helpers.getYesterday()));
          }
          getOSLaunchesData(db, argv.coll, numDays, callback);
        });
      });
    },
    getVersionLaunchesData,
    getSessionData,
    getSamplingData,
    getInstanceData,
    getUpdateData,
    writeObj,
    writeToDB
  ], function(error, results) {
    if (error) {
      console.error(error);
    }
    console.log(results);
  });
}

exports.command = 'mongo [url] [db] [coll]';
exports.desc = 'Export report data to MongoDB';
exports.builder = {
  url: {
    default: 'mongodb://localhost:27017/'
  },
  db: {
    default: 'test'
  },
  coll: {
    default: 'compass'
  }
};
exports.handler = function(argv) {
  exportToMongo(argv);
};
