/**
 * This is the file that handles the production of the command line version
 * of the Compass Usage Report. The program makes use of the async.waterfall()
 * command so as to collect each of the metrics and use them in synchronous
 * fashion. Most of the data is presented in table format, which is created
 * by the module cli-table. The function to handle the creation of the table,
 * makeTable(), exists in helpers.js. Each function in the waterfall gets the
 * data in an appropriate format to pass to makeTable or to be printed to the
 * console.
*/

/* eslint no-console:0 */

require('dotenv').config();
var usageReport = require('../../');
var async = require('async');
var chalk = require('chalk');
var helpers = require('../../lib/helpers.js');

var options = require('yargs').argv;

// ensure the .env file is properly configured
helpers.verifyEnv();

/**
 * Function to fetch the metric for number of app launches by OS per day.
 * Passes table of data to the callback.
 * @param {function} callback        callback to pass the data through the waterfall
*/
function asyncOsLaunches(callback) {
  var osLaunchesDataArray = [];
  var stream = usageReport.getLaunchesOsData(process.env.MIXPANEL_API_SECRET);
  stream.on('data', function(data) {
    osLaunchesDataArray.push({
      date: data.date,
      'Mac OS X': data.values['Mac OS X'],
      'Windows': data.values.Windows
    });
  });
  stream.on('end', function() {
    var osLaunchesTable = helpers.makeTable(['date', 'Mac OS X', 'Windows'], osLaunchesDataArray);
    callback(null, osLaunchesTable.toString());
  });
}

/**
 * Function to fetch the metric for number of app launches by Compass version per day.
 * Passes table of data to the callback.
 * @param {table} osLaunches         table of osLaunches data to be passed through waterfall
 * @param {function} callback        callback to pass all data through the waterfall
*/
function asyncVersionLaunches(osLaunches, callback) {
  var versionLaunchesDataArray = [];
  var stream = usageReport.getLaunchesVersionData(process.env.MIXPANEL_API_SECRET);
  var versions = [];
  stream.on('data', function(data) {
    var dataObj = {};
    dataObj.date = data.date;
    for (var vers in data.values) {
      if (data.values.hasOwnProperty(vers)) {
        dataObj[vers] = data.values[vers];
      }
      if (!helpers.isInArray(vers, versions)) {
        versions.push(vers);
      }
    }
    versionLaunchesDataArray.push(dataObj);
  });
  stream.on('end', function() {
    versions.sort().unshift('date');
    var versionLaunchesTable = helpers.makeTable(versions, versionLaunchesDataArray);
    callback(null, osLaunches, versionLaunchesTable.toString());
  });
}

/**
 * Function to fetch the metric for sessions per user per day. Each data
 * event emitted looks like:
 * { date: '2016-08-02',
 *   values: {
 *     user1: {
 *       sessions: 1,
 *       lengths: 150
 *     },
 *     user2: {
 *       sessions: 2,
 *       lengths: 6
 *     }
 *   }
 * }
 * This function transforms the data to compute statistics for number of sessions
 * per user per day in addition to length of sessions per user.
 * Passes table of data to the callback.
 * @param {table} osLaunches         table of osLaunches data to be passed through waterfall
 * @param {table} versionLaunches    table of versionLaunches data to be passed through waterfall
 * @param {function} callback        callback to pass all data through the waterfall
*/
function asyncSessionData(osLaunches, versionLaunches, callback) {
  var stream = usageReport.getSessionData(process.env.MIXPANEL_API_SECRET);
  var userDict = {};
  stream.on('data', function(chunk) {
    for (var user in chunk.values) {
      if (chunk.values.hasOwnProperty(user)) {
        if (userDict[user] === undefined) {
          userDict[user] = {
            days: 1,
            sessions: chunk.values[user].sessions,
            lengths: chunk.values[user].lengths
          };
        } else {
          userDict[user].days++;
          userDict[user].sessions += chunk.values[user].sessions;
          userDict[user].lengths += chunk.values[user].lengths;
        }
      }
    }
  });
  stream.on('end', function() {
    var dataObj = {
      days: {
        property: 'days',
        values: []
      },
      sessions: {
        property: 'sessions',
        values: []
      },
      lengths: {
        property: 'lengths',
        values: []
      }
    };
    for (var user in userDict) {
      if (userDict.hasOwnProperty(user)) {
        dataObj.days.values.push(userDict[user].days);
        dataObj.sessions.values.push(userDict[user].sessions);
        dataObj.lengths.values.push(userDict[user].lengths);
      }
    }
    var sessionDataArray = [];
    for (var key in dataObj) {
      if (dataObj.hasOwnProperty(key)) {
        var row = helpers.getStats(dataObj[key].values);
        row.property = dataObj[key].property;
        sessionDataArray.push(row);
      }
    }
    var table = helpers.makeTable(['property', 'max', 'avg', 'med', 25, 75], sessionDataArray);
    callback(null, osLaunches, versionLaunches, table.toString());
  });
}

/**
 * Function to gather data on sampling times. Each data event contains every
 * single sample time for a given data. Since we are looking at stats on sampling
 * time over the last month, this function aggregates the data for each day into
 * a single array of all sample times for the month.
 * Passes table of data to the callback.
 * @param {table} osLaunches         table of osLaunches data to be passed through waterfall
 * @param {table} versionLaunches    table of versionLaunches data to be passed through waterfall
 * @param {table} session            table of session data to be passed through waterfall
 * @param {function} callback        callback to pass all data through the waterfall
*/
function asyncSamplingData(osLaunches, versionLaunches, session, callback) {
  var stream = usageReport.getSamplingData(process.env.MIXPANEL_API_SECRET);
  var samplingDataArray = [];
  stream.on('data', function(data) {
    samplingDataArray = samplingDataArray.concat(data.values);
  });
  stream.on('end', function() {
    var dataObj = helpers.getStats(samplingDataArray);
    dataObj.property = 'sampling';
    var table = helpers.makeTable(['property', 'max', 'avg', 'med', 25, 75], [dataObj]);
    callback(null, osLaunches, versionLaunches, session, table.toString());
  });
}

/**
 * Function to get data on app connection conversion. This data is fetched from
 * the funnel endpoint of the Mixpanel API. The resulting table shows each stepLen
 * and the conversion rate for each step. Since the data is returned on a per days
 * basis, the data is first aggregated before the stats are compiled into a table.
 * Passes table of data to the callback.
 * @param {table} osLaunches         table of osLaunches data to be passed through waterfall
 * @param {table} versionLaunches    table of versionLaunches data to be passed through waterfall
 * @param {table} session            table of session data to be passed through waterfall
 * @param {table} sampling           table of sampling data to be passed through waterfall
 * @param {function} callback        callback to pass all data through the waterfall
*/
function asyncInstanceData(osLaunches, versionLaunches, session, sampling, callback) {
  var instanceConnectionDataArray = [];
  var stream = usageReport.getInstanceData(process.env.MIXPANEL_API_SECRET);
  var dataObj = {};
  var i;
  stream.on('data', function(data) {
    if (dataObj.analysis === undefined) {
      dataObj.analysis = data.values.analysis;
      dataObj.steps = data.values.steps;
    } else {
      dataObj.analysis.starting_amount += data.values.analysis.starting_amount;
      dataObj.analysis.completion += data.values.analysis.completion;
      for (i = 0; i < data.values.steps.length; ++i) {
        dataObj.steps[i].count += data.values.steps[i].count;
      }
    }
  });
  stream.on('end', function() {
    var started = dataObj.steps[0].count;
    for (i = 0; i < dataObj.steps.length; ++i) {
      instanceConnectionDataArray.push({
        step: dataObj.steps[i].event,
        started: started,
        completed: dataObj.steps[i].count,
        conversion: (dataObj.steps[i].count / started).toFixed(2)
      });
      started = dataObj.steps[i].count;
    }
    instanceConnectionDataArray.push({
      step: 'total',
      started: dataObj.analysis.starting_amount,
      completed: dataObj.analysis.completion,
      conversion: (dataObj.analysis.completion / dataObj.analysis.starting_amount).toFixed(2)
    });
    var instanceConnectionTable = helpers.makeTable(['step', 'started', 'completed', 'conversion'], instanceConnectionDataArray);
    callback(null, osLaunches, versionLaunches, session, sampling, instanceConnectionTable.toString());
  });
}

/**
 * Function to get data on auto update conversion. This data is fetched from
 * the funnel endpoint of the Mixpanel API. The resulting table shows each stepLen
 * and the conversion rate for each step. Since the data is returned on a per days
 * basis, the data is first aggregated before the stats are compiled into a table.
 * Passes table of data to the callback.
 * @param {table} osLaunches         table of osLaunches data to be passed through waterfall
 * @param {table} versionLaunches    table of versionLaunches data to be passed through waterfall
 * @param {table} session            table of session data to be passed through waterfall
 * @param {table} sampling           table of sampling data to be passed through waterfall
 * @param {table} instance           table of instance connection data to be passed through waterfall
 * @param {function} callback        callback to pass all data through the waterfall
*/
function asyncUpdateData(osLaunches, versionLaunches, session, sampling, instance, callback) {
  var updateDataArray = [];
  var stream = usageReport.updateSuccessData(process.env.MIXPANEL_API_SECRET);
  var dataObj = {};
  var i;
  stream.on('data', function(data) {
    if (dataObj.analysis === undefined) {
      dataObj.analysis = data.values.analysis;
      dataObj.steps = data.values.steps;
    } else {
      dataObj.analysis.starting_amount += data.values.analysis.starting_amount;
      dataObj.analysis.completion += data.values.analysis.completion;
      for (i = 0; i < data.values.steps.length; ++i) {
        dataObj.steps[i].count += data.values.steps[i].count;
      }
    }
  });
  stream.on('end', function() {
    var started = dataObj.steps[0].count;
    for (i = 0; i < dataObj.steps.length; ++i) {
      updateDataArray.push({
        step: dataObj.steps[i].event,
        started: started,
        completed: dataObj.steps[i].count,
        conversion: (dataObj.steps[i].count / started).toFixed(2)
      });
      started = dataObj.steps[i].count;
    }
    updateDataArray.push({
      step: 'total',
      started: dataObj.analysis.starting_amount,
      completed: dataObj.analysis.completion,
      conversion: (dataObj.analysis.completion / dataObj.analysis.starting_amount).toFixed(2)
    });
    var updateDataTable = helpers.makeTable(['step', 'started', 'completed', 'conversion'], updateDataArray);
    callback(null, osLaunches, versionLaunches, session, sampling, instance, updateDataTable.toString());
  });
}

/**
 * Function to get data on an active user breakdown by OS. This data currently
 * comes from Intercom, but starting with version 1.4.0-beta, we should start looking
 * into getting this data from Mixpanel to speed up the report time. The function
 * compiles a readable string (using new lines and tabs) to print to the console.
 * @param {table} osLaunches         table of osLaunches data to be passed through waterfall
 * @param {table} versionLaunches    table of versionLaunches data to be passed through waterfall
 * @param {table} session            table of session data to be passed through waterfall
 * @param {table} sampling           table of sampling data to be passed through waterfall
 * @param {table} instance           table of instance connection data to be passed through waterfall
 * @param {table} update             table of auto update flow data to be passed through waterfall
 * @param {function} callback        callback to pass all data through the waterfall
*/
function asyncUsersData(osLaunches, versionLaunches, session, sampling, instance, update, callback) {
  var outputString = '';
  var mainOS = ['Total', 'Mac', 'Windows', 'Linux', 'undefined'];
  var promise = usageReport.getActiveUsers();
  promise.then(function(data) {
    for (var key in data) {
      if (data.hasOwnProperty(key)) {
        if (!helpers.isInArray(key, mainOS)) {
          outputString += '\t';
        }
        outputString += chalk.green.underline.underline(key) + ': ' + data[key] + '\n';
      }
    }
    callback(null, osLaunches, versionLaunches, session, sampling, instance, update, outputString);
  }, function(reason) {
    console.log(reason);
  });
}

/**
 * Function to execute the waterfall. Grabs all the data and then synchronously
 * prints it to the console.
*/
function produceReport() {
  async.waterfall([
    asyncOsLaunches,
    asyncVersionLaunches,
    asyncSessionData,
    asyncSamplingData,
    asyncInstanceData,
    asyncUpdateData,
    asyncUsersData,
    function(osLaunches, versionLaunches, session, sampling, instance, update, users) {
      if (options.html) {
        console.log('<html>\n<pre>');
      }
      console.log(chalk.inverse.bold('App Launches By OS'));
      console.log(osLaunches);
      console.log();
      console.log(chalk.inverse.bold('App Launches By Compass Version'));
      console.log(versionLaunches);
      console.log();
      console.log(chalk.inverse.bold('Session (min)'));
      console.log(session);
      console.log();
      console.log(chalk.inverse.bold('Sampling (ms)'));
      console.log(sampling);
      console.log();
      console.log(chalk.inverse.bold('Instance Connection Flow'));
      console.log(instance);
      console.log();
      console.log(chalk.inverse.bold('Update Connection Successful'));
      console.log(update);
      console.log();
      console.log(chalk.inverse.bold('Users'));
      console.log(users);
      if (options.html) {
        console.log('</pre>\n</html>');
      }
    }
  ]);
}

exports.command = 'cli';
exports.desc = 'Produce a command line report';
exports.builder = {};
exports.handler = produceReport;
