var _ = require('highland');
var Table = require('cli-table');

/**
 * Function to ensure all of the write variable are configured in the
 * .env file.
*/
function verifyEnv() {
  if (! (process.env.INTERCOM_APP_ID &&
          process.env.INTERCOM_API_KEY &&
          process.env.updateSuccessID &&
          process.env.updateCancelID &&
          process.env.instanceID &&
          process.env.MIXPANEL_API_SECRET)) {
    throw new Error('Improperly configured .env file: not the right environment variables');
  }
}

/**
 * Function that takes in a Date Object and converts it into
 * the format 'YYYY-MM-DD'
 * @param {Date} date        date object to be formatted
 * @returns {string}         formatted date
*/
function formatDate(date) {
  var year = date.getFullYear().toString();
  var month = (date.getMonth() + 1).toString();
  var mm = month[1] ? month : '0' + month;
  var day = date.getDate().toString();
  var dd = day[1] ? day : '0' + day;
  return year + '-' + mm + '-' + dd;
}

/**
 * Function to sort such that the least value comes first.
 * Passed into the .sort() function as a parameter.
 * @param {int} a          for sorting
 * @param {int} b          for sorting
 * @returns {int}          for sorting
*/
function sortAscending(a, b) {
  return a - b;
}

/**
 * Function to get today's date and return it as a date object.
 * @returns {Date}          today's date
*/
function getToday() {
  return new Date();
}

/**
 * Function to get the date exactly a month ago and return it as a date object.
 * @returns {Date}          date from one month ago
*/
function getMonthAgo() {
  return new Date((new Date()).setMonth((new Date()).getMonth() - 1));
}

/**
 * Function to get the yesterday's date and return it as a date object.
 * @returns {Date}          yesterday's date
*/
function getYesterday() {
  return new Date((new Date()).setDate((new Date()).getDate() - 1));
}

/**
 * Function to sort dates such that the most recent date comes first.
 * Passed as a parameter to the .sort() function.
 * @param {date} date1          for sorting
 * @param {date} date2          for sorting
 * @returns {int}               for sorting
*/
function sortDates(date1, date2) {
  var utc1 = new Date(date1);
  var utc2 = new Date(date2);
  return utc2 - utc1;
}

/**
 * Function to sum up the values in an array.
 * @param {array} arr       array whose values are summed
 * @returns {int}             sum of values in array
*/
function sum(arr) {
  var len = arr.length;
  var total = 0;
  for (var i = 0; i < len; ++i) {
    total += arr[i];
  }
  return total;
}

/**
 * Function to get the max, median, avg, 25th, and 75th percentiles
 * of an array of numbers.
 * @param {array} arr         array for which the function calculates statistics
 * @returns {obj}             object with stats
*/
function getStats(arr) {
  arr.sort(sortAscending);
  var total = sum(arr);
  var obj = {
    max: arr[arr.length - 1],
    avg: (total / arr.length).toFixed(0),
    med: arr[Math.round(arr.length * 0.5)],
    25: arr[Math.round(arr.length * 0.25)],
    75: arr[Math.round(arr.length * 0.75)]
  };
  return obj;
}

/**
 * Function to transform the object returned from a JQL query and
 * create a stream of objects, where each object represents a days'
 * worth of data.
 * Object comes in the form:
 * {
 *    2016-08-02: data
 *    2016-08-01: data
 * }
 * Gets transformed such that each data event emitted has form:
 * {
 *  date: 2016-08-02,
 *  values: data
 * }
 * @param {stream} str          stream of data returned from mixpanel.mixpanelJQL
 * @returns {stream}            stream of data with events emitted per day
*/
function transformStreamJQL(str) {
  return _(function(push) {
    var data;
    str.on('data', function(chunk) {
      data = chunk;
    });
    str.on('end', function() {
      var dates = Object.keys(data);
      dates.sort(sortDates);
      var len = dates.length;
      for (var i = 0; i < len; ++i) {
        push(null, {
          date: dates[i],
          values: data[dates[i]]
        });
      }
      push(null, _.nil);
    });
  });
}

/**
 * Function that transforms streams coming in from the funnels endpoint.
 * Data is returned in the form:
 * { meta: {
 *      dates: ['2016-08-02', '2016-08-01']
 *    }
 *   data: {
 *      '2016-08-02': stuff,
 *      '2016-08-01': more stuff
 *    }
 * }
 * Function transforms data into:
 * {
 *  date: 2016-08-02,
 *  values: stuff
 * }
 * @param {stream} str         stream of data returned from mixpanel.funnel
 * @returns {stream}            stream of data with events emitted per day
*/
function transformStreamFunnel(str) {
  return _(function(push) {
    str.on('data', function(chunk) {
      var len = chunk.meta.dates.length;
      var date;
      for (var i = len - 1; i >= 0; --i) {
        date = chunk.meta.dates[i];
        var obj = {};
        obj.date = date;
        obj.values = {};
        obj.values.analysis = {
          starting_amount: chunk.data[date].analysis.starting_amount,
          completion: chunk.data[date].analysis.completion
        };
        obj.values.steps = [];
        var stepLen = chunk.data[date].steps.length;
        for (var j = 0; j < stepLen; ++j) {
          obj.values.steps.push({
            event: chunk.data[date].steps[j].event,
            count: chunk.data[date].steps[j].count
          });
        }
        push(null, obj);
      }
      push(null, _.nil);
    });
  });
}

/**
 * Function that makes a cli-table table given an array of headers (keys)
 * and an array of objects, each object being a row in the table.
 * Signature of parameters:
 * keys: [key1, key2, key3]
 * arr: [{
 *          key1: something,
 *          key2: something else
 *          key3: something else again
 *        },
 *       {
 *          key1: data
 *          key2: more data
 *          key3: even more data
 *         }]
 * @param {array} keys            list of headers for the table
 * @param {array} arr             list of objects that will comprise rows of table
 * @returns {Table}               cli-table returned - invoked with toString()
*/
function makeTable(keys, arr) {
  var table = new Table({
    head: keys,
    style: {
      head: ['green']
    }
  });
  for (var i = 0; i < arr.length; ++i) {
    var newRow = [];
    for (var j = 0; j < keys.length; ++j) {
      newRow.push(arr[i][keys[j]] || 0);
    }
    table.push(newRow);
  }
  return table;
}

/**
 * Function to check if a value is in an array.
 * @param {int|string|char|bool} value      needle
 * @param {array} array                     haystack
 * @returns {boolean}                       true if value is in array
*/
function isInArray(value, array) {
  return array.indexOf(value) > -1;
}

/**
 * Function to normalize dates - used to calculate number of days between
 * two date objects.
 * @param {date} date           date object to be treated as UTC
 * @returns {int}               new UTC date with offset accounted for
*/
function treatAsUTC(date) {
  var result = new Date(date);
  result.setMinutes(result.getMinutes() - result.getTimezoneOffset());
  return result;
}

/**
 * Function to get difference in days between two dates.
 * @param {date} startDate           start date of the two days
 * @param {date} endDate             end date of the two days
 * @returns {int}                    number of days between start date and end date
*/
function daysBetween(startDate, endDate) {
  var millisecondsPerDay = 24 * 60 * 60 * 1000;
  return (treatAsUTC(endDate) - treatAsUTC(startDate)) / millisecondsPerDay;
}

function getURL(url, db) {
  if (!db) {
    return url;
  } else if (url[url.length - 1] === '/') {
    return url + db;
  }
  return url + '/' + db;
}

module.exports = {
  formatDate: formatDate,
  getToday: getToday,
  getYesterday: getYesterday,
  getMonthAgo: getMonthAgo,
  sortAscending: sortAscending,
  sortDates: sortDates,
  transformStreamJQL: transformStreamJQL,
  transformStreamFunnel: transformStreamFunnel,
  sum: sum,
  getStats: getStats,
  monthInterval: 32,
  makeTable: makeTable,
  isInArray: isInArray,
  verifyEnv: verifyEnv,
  daysBetween: daysBetween,
  getURL: getURL
};
