/* eslint-env es6 */

var mixpanel = require('mixpanel-api-wrapper');
var helpers = require('./helpers.js');

var yesterday = helpers.getYesterday();
var monthAgo = helpers.getMonthAgo();

/**
 * JQL query that grabs data about sampling. Essentially queries for all
 * sample events over the course of a month and returns a list of sample times
 * for each day. Object returned is in the form:
 * {
 *  '2016-08-02': [1, 2, 3, 4],
 *  '2016-08-01': [40, 22, 3, 410]
 * }
*/
var samplingQuery = `function main() {
  function getDay(event) {
    return (new Date(event.time)).toISOString().split('T')[0];
  }
  return Events({
    from_date: '${helpers.formatDate(monthAgo)}',
    to_date: '${helpers.formatDate(yesterday)}'
  })
  .filter(function(event) {
    return event.name === "Schema sampled";
  })
  .filter(function(event) {
    return /^((?!(dev|world|testing)).)*$/.test(event.properties['app version']);
  })
  .map(function(item) {
    var sampleTime = item.properties['total sample time'] + item.properties['total analysis time'];
    return {
      'date': getDay(item),
      'sampleTime': sampleTime
    };
  })
  .reduce(function(accumulators, items) {
    var obj = {};
    var i;
    for (i = 0; i < accumulators.length; ++i) {
      for (var date in accumulators[i]) {
        if (obj[date] === undefined) {
          obj[date] = [];
        }
        for (var j = 0; j < accumulators[i][date].length; ++j) {
          obj[date].push(accumulators[i][date][j]);
        }
      }
    }
    for (i = 0; i < items.length; ++i) {
      if (obj[items[i].date] === undefined) {
        obj[items[i].date] = [];
      }
      obj[items[i].date].push(items[i].sampleTime);
    }
    return obj;
  });
}`;

/**
 * Function that gets sampling data. Bulk of the
 * logic occurs in above JQL query.
 * @param {string} mixpanelAPISecret           Mixpanel API secret to access team's data
 * @return {stream}                 returns stream of results from Mixpanel API
*/
function getSamplingData(mixpanelAPISecret) {
  var stream = mixpanel.mixpanelJQL(mixpanelAPISecret, samplingQuery);
  var transformedStream = helpers.transformStreamJQL(stream);
  return transformedStream;
}

module.exports = getSamplingData;
