/* eslint-env es6 */

var mixpanel = require('mixpanel-api-wrapper');
var helpers = require('./helpers.js');

var yesterday = helpers.getYesterday();

var monthAgo = helpers.getMonthAgo();

/**
 * JQL query to grab session data per user. Object returned in the form:
 * {
 *  '2016-08-02': {
 *        user1: {sessions: 2, lengths: 10},
 *        user2: {sessions: 4, lengths: 62}
 *    },
 *  '2016-08-01': {
 *        user2: {sessions: 1, lengths: 7},
 *        user3: {sessions: 2, lengths: 18}
 *    }
 * }
 * That is, for a given day, there is data for every user who used Compass.
 * The return object details how many sessions a user used Compass in the day
 * and also the total length of time during which they used Compass.
*/
var sessionsPerUser = `function main() {
  function getDay(event) {
    return (new Date(event.time)).toISOString().split('T')[0];
  }
  return Events({
    from_date: '${helpers.formatDate(monthAgo)}',
    to_date: '${helpers.formatDate(yesterday)}'
  })
  .filter(function(event) {
    return event.name === "App quit";
  })
  .filter(function(event) {
    return /^((?!(dev|world|testing)).)*$/.test(event.properties['app version']);
  })
  .map(function(event) {
    var obj = {
      id: event.distinct_id,
      length: event.properties['minutes since start'],
      date: getDay(event)
    };
    return obj;
  })
  .reduce(function(accumulators, items) {
    var i;
    var obj = {};
    for(i = 0; i < accumulators.length; ++i) {
      for (var d in accumulators[i]) {
        if (obj[d] === undefined) {
          obj[d] = {};
        }
        for (var u in accumulators[i][d]) {
          if (obj[d][u] === undefined) {
            obj[d][u] = {
              sessions: accumulators[i][d][u].sessions,
              lengths: accumulators[i][d][u].lengths
            };
          }
          else {
            obj[d][u].sessions += accumulators[i][d][u].sessions;
            obj[d][u].lengths += accumulators[i][d][u].lengths;
          }
        }
      }
    }
    for(i = 0; i < items.length; ++i) {
      var date = items[i].date;
      var id = items[i].id;
      var length = items[i].length;
      if (obj[date] === undefined) {
        obj[date] = {};
      }
      if (obj[date][id] === undefined) {
        obj[date][id] = {
          sessions: 1,
          lengths: length
        };
      }
      else {
        obj[date][id].sessions += 1;
        obj[date][id].sessions += length;
      }
    }
    return obj;
  });
}`;

/**
 * Function that gets session data per day per user. Bulk of the
 * logic occurs in above JQL query.
 * @param {string} mixpanelAPISecret           Mixpanel API secret to access team's data
 * @return {stream}                 returns stream of results from Mixpanel API
*/
function getSessionsPerUser(mixpanelAPISecret) {
  var stream = mixpanel.mixpanelJQL(mixpanelAPISecret, sessionsPerUser);
  var transformedStream = helpers.transformStreamJQL(stream);
  return transformedStream;
}

module.exports = getSessionsPerUser;
